@@ -6,7 +6,6 @@ from formsite_util.core import FormsiteParams, FormsiteCredentials, FormsiteInte
 from time import perf_counter
 
 
-
 def GatherArguments():
     parser = argparse.ArgumentParser(
         description="Github of author: https://github.com/strny0/formsite-utility\n"
@@ -26,7 +25,7 @@ def GatherArguments():
                     "| 5xx  | Unexpected internal error.                  |\n",
                     formatter_class=argparse.RawTextHelpFormatter
     )
-    parser.add_argument('-V', '--version', action='version',version="1.2.6.5")
+    parser.add_argument('-V', '--version', action='version', version="1.2.6.6")
     g_login = parser.add_argument_group('Login')
     g_params = parser.add_argument_group('Results Parameters')
     g_functions = parser.add_argument_group('Functions')
@@ -136,6 +135,12 @@ def GatherArguments():
                                "\nAny files that would be overwritten as a result of the removal of characters will be appended with _1, _2, etc.")
     g_functions.add_argument('-S', '--store_latest_ref',  nargs='?',  default=False, const='default',
                              help="If you enable this option, a text file `latest_ref.txt` will be created. \nThis file will only contain the highest reference number in the export. \nIf there are no results in your export, nothign will happen.")
+    g_functions.add_argument('--timeout',  nargs=1,  default=30, type=int,
+                             help="Timeout in seconds for when you are downloading files. Defaults to 30.")
+    g_functions.add_argument('--retries',  nargs=1,  default=3, type=int,
+                             help="Number of times to retry downloading files if the download fails. Defaults to 3.")
+    g_functions.add_argument('--get_download_status', default=False, action='store_true',
+                             help="If you enable this option, a text file with status for each downloaded link will be saved (complete or incomplete).")
     g_nocreds.add_argument('-l', '--list_columns', action="store_true",  default=False,
                            help="If you use this flag, program will output mapping of what column belongs to which column ID instead of actually running, useful for figuring out search arguments."
                            "\nRequires login info and form id. Overrides all other functionality of the program."
@@ -216,7 +221,7 @@ def main():
                     default_folder, max_concurrent_downloads=arguments.concurrent_downloads, links_regex=arguments.links_regex, filename_regex=arguments.filename_regex, overwrite_existing=arguments.dont_overwrite_downloads)
             else:
                 interface.DownloadFiles(
-                    arguments.download_links, max_concurrent_downloads=arguments.concurrent_downloads, links_regex=arguments.links_regex, filename_regex=arguments.filename_regex, overwrite_existing=arguments.dont_overwrite_downloads)
+                    arguments.download_links, max_concurrent_downloads=arguments.concurrent_downloads, links_regex=arguments.links_regex, filename_regex=arguments.filename_regex, overwrite_existing=arguments.dont_overwrite_downloads, report_downloads=arguments.get_download_status, timeout=arguments.timeout, retries=arguments.retries)
             print("download complete")
         if arguments.store_latest_ref is not False:
             if arguments.store_latest_ref == 'default':
